package edu.illinois.cs.cs124.ay2023.mp.test

import android.content.Intent
import androidx.test.core.app.ApplicationProvider
import androidx.test.espresso.Espresso.onView
import androidx.test.espresso.action.ViewActions.click
import androidx.test.espresso.assertion.ViewAssertions.matches
import androidx.test.espresso.matcher.ViewMatchers.hasDescendant
import androidx.test.espresso.matcher.ViewMatchers.isDisplayed
import androidx.test.espresso.matcher.ViewMatchers.withId
import androidx.test.espresso.matcher.ViewMatchers.withSubstring
import androidx.test.espresso.matcher.ViewMatchers.withText
import androidx.test.ext.junit.runners.AndroidJUnit4
import com.fasterxml.jackson.module.kotlin.readValue
import com.google.common.truth.Truth.assertWithMessage
import edu.illinois.cs.cs124.ay2023.mp.R
import edu.illinois.cs.cs124.ay2023.mp.activities.CourseActivity
import edu.illinois.cs.cs124.ay2023.mp.models.Course
import edu.illinois.cs.cs124.ay2023.mp.models.Summary
import edu.illinois.cs.cs124.ay2023.mp.network.Client
import edu.illinois.cs.cs124.ay2023.mp.network.Server
import edu.illinois.cs.cs124.ay2023.mp.network.startServer
import edu.illinois.cs.cs124.ay2023.mp.test.helpers.COURSES
import edu.illinois.cs.cs124.ay2023.mp.test.helpers.CoursesReadCountSecurityManager
import edu.illinois.cs.cs124.ay2023.mp.test.helpers.GET_METHOD_EXTRA_TIME
import edu.illinois.cs.cs124.ay2023.mp.test.helpers.SUMMARIES
import edu.illinois.cs.cs124.ay2023.mp.test.helpers.SUMMARY_COUNT
import edu.illinois.cs.cs124.ay2023.mp.test.helpers.compareCourses
import edu.illinois.cs.cs124.ay2023.mp.test.helpers.configureLogging
import edu.illinois.cs.cs124.ay2023.mp.test.helpers.countRecyclerView
import edu.illinois.cs.cs124.ay2023.mp.test.helpers.objectMapper
import edu.illinois.cs.cs124.ay2023.mp.test.helpers.pause
import edu.illinois.cs.cs124.ay2023.mp.test.helpers.startActivity
import edu.illinois.cs.cs124.ay2023.mp.test.helpers.startMainActivity
import edu.illinois.cs.cs124.ay2023.mp.test.helpers.testClient
import edu.illinois.cs.cs124.ay2023.mp.test.helpers.testServerGet
import edu.illinois.cs.cs124.ay2023.mp.test.helpers.testServerGetTimed
import edu.illinois.cs.cs124.ay2023.mp.test.helpers.toPath
import edu.illinois.cs.cs124.ay2023.mp.test.helpers.withRecyclerView
import edu.illinois.cs.cs125.gradlegrader.annotations.Graded
import org.junit.After
import org.junit.AfterClass
import org.junit.Before
import org.junit.BeforeClass
import org.junit.FixMethodOrder
import org.junit.Test
import org.junit.runner.RunWith
import org.junit.runners.MethodSorters
import org.robolectric.Shadows
import org.robolectric.annotation.LooperMode
import org.robolectric.annotation.experimental.LazyApplication
import java.lang.reflect.Modifier
import java.net.HttpURLConnection
import kotlin.math.floor
import kotlin.random.Random

/*
 * This is the MP2 test suite.
 * The code below is used to evaluate your app during testing, local grading, and official grading.
 * You may not understand all of the code below, but you'll need to have some understanding of how
 * it works so that you can determine what is wrong with your app and what you need to fix.
 *
 * ALL CHANGES TO THIS FILE WILL BE OVERWRITTEN DURING OFFICIAL GRADING.
 * You can and should modify the code below if it is useful during your own local testing,
 * but any changes you make will be discarded during official grading.
 * The local grader will not run if the test suites have been modified, so you'll need to undo any
 * local changes before you run the grader.
 *
 * Note that this means that you should not fix problems with the app by modifying the test suites.
 * The test suites are always considered to be correct.
 *
 * Our test suites are broken into two parts.
 * The unit tests are tests that we can perform without running your app.
 * They test things like whether a method works properly or the behavior of your API server.
 * Unit tests are usually fairly fast.
 *
 * The integration tests are tests that require simulating your app.
 * This allows us to test things like your API client, and higher-level aspects of your app's
 * behavior, such as whether it displays the right thing on the display.
 * Because integration tests require simulating your app, they run more slowly.
 *
 * The MP2 test suite includes no ungraded tests.
 * These tests are fairly idiomatic, in that they resemble tests you might write for an actual
 * Android programming project.
 */

@RunWith(AndroidJUnit4::class)
@LooperMode(LooperMode.Mode.PAUSED)
@FixMethodOrder(MethodSorters.NAME_ASCENDING)
class MP2Test {
    // /////////////////////////////////////////////////////////////////////////////////////////////
    // Unit tests that don't require simulating the entire app, and usually complete quickly
    // /////////////////////////////////////////////////////////////////////////////////////////////

    /** Test the Course class. */
    @Test(timeout = 1000L)
    @Graded(points = 10, friendlyName = "Course Class Design")
    @LazyApplication(LazyApplication.LazyLoad.ON)
    fun test0_CourseClassDesign() {
        for (expectedString in COURSES) {
            val course: Course = objectMapper.readValue(expectedString)
            val courseString: String = objectMapper.writeValueAsString(course)
            compareCourses(expectedString, courseString)
        }
    }

    // Remove outliers for more stable mean
    private fun List<Long>.trimmedMean(percent: Double = 0.1): Double {
        val trimmedList = toMutableList().sorted().let {
            val toDrop = floor(size * percent).toInt()
            subList(toDrop, size - toDrop)
        }.toList()
        return trimmedList.sum().toDouble() / trimmedList.size
    }

    /** Test GET /course/ server route. */
    @Test(timeout = 10000L)
    @Graded(points = 20, friendlyName = "Server GET /course/")
    @LazyApplication(LazyApplication.LazyLoad.ON)
    fun test1_ServerCourseRoute() {
        // Save response times for comparison
        val baseResponseTimes = mutableListOf<Long>()
        val getResponseTimes = mutableListOf<Long>()

        // Test good GET /course/ requests for all courses
        for (expectedString in COURSES) {
            val baseResult = "/".testServerGetTimed<String>(String::class.java)
            baseResponseTimes += baseResult.responseTime.toNanos()

            val node = objectMapper.readTree(expectedString)!!
            val courseResult = "/course/${node.toPath()}".testServerGetTimed<Course>(Course::class.java)
            val courseString: String = objectMapper.writeValueAsString(courseResult.response)
            compareCourses(expectedString, courseString)
            getResponseTimes += courseResult.responseTime.toNanos()
        }

        // Check for slow server GETs potentially caused by unnecessary parsing or looping
        val averageBase = baseResponseTimes.trimmedMean()
        val averageResponse = getResponseTimes.trimmedMean()
        assertWithMessage("Server GET /course/ is too slow")
            .that(averageResponse)
            .isLessThan(averageBase * GET_METHOD_EXTRA_TIME)

        // Test bad requests
        // Bad URL
        "/course/CS/".testServerGet(HttpURLConnection.HTTP_BAD_REQUEST)
        // Non-existent course
        "/course/CS/188".testServerGet(HttpURLConnection.HTTP_NOT_FOUND)
        // Non-existent URL (trailing slash)
        "/courses/CS/124/".testServerGet(HttpURLConnection.HTTP_NOT_FOUND)
    }

    // /////////////////////////////////////////////////////////////////////////////////////////////
    // Integration tests that require simulating the entire app, and are usually slower
    // /////////////////////////////////////////////////////////////////////////////////////////////

    // Indices of summaries to use for the UI test
    private val summariesForUITest = (0 until SUMMARY_COUNT).shuffled(Random(124)).take(4)

    /** Test the client getCourse method. */
    @Test(timeout = 20000L)
    @Graded(points = 10, friendlyName = "Client GET /course/")
    @LazyApplication(LazyApplication.LazyLoad.ON)
    fun test2_ClientGetCourse() {
        // Test getCourse requests for most courses
        for (i in 0 until SUMMARY_COUNT) {
            // Skip course summaries we use for the UI test
            if (i in summariesForUITest) {
                continue
            }
            val summary = SUMMARIES[i]
            val expectedString = COURSES[i]
            val course: Course = testClient { callback -> Client.getCourse(summary, callback) }
            compareCourses(expectedString, objectMapper.writeValueAsString(course))
        }
        try {
            testClient { callback -> Client.getCourse(Summary("Not", "A", "Course"), callback) }
            assertWithMessage("Client GET /course/ for non-existent course should throw").fail()
        } catch (_: Exception) {
        }
    }

    /** Test onClick intent generation in MainActivity. */
    @Test(timeout = 10000L)
    @Graded(points = 20, friendlyName = "Summary Click Launch")
    @LazyApplication(LazyApplication.LazyLoad.ON)
    fun test3_SummaryClickLaunch() {
        // Launch the main activity and confirm correct transition to CourseActivity
        startMainActivity { activity ->
            // Sanity checks
            onView(withId(R.id.recycler_view)).check(countRecyclerView(SUMMARY_COUNT))
            onView(withRecyclerView(R.id.recycler_view).atPosition(0))
                .check(matches(hasDescendant(withText("CS 100: Computer Science Orientation"))))

            // Perform the click
            onView(withRecyclerView(R.id.recycler_view).atPosition(0)).perform(click())

            // Make sure the intent generated is correct
            val courseExtra = Shadows.shadowOf(activity).nextStartedActivity.getStringExtra("summary")
            val node = objectMapper.readTree(courseExtra)
            assertWithMessage("Intent contains incorrect subject")
                .that(node["subject"].asText())
                .isEqualTo("CS")
            assertWithMessage("Intent contains incorrect number")
                .that(node["number"].asText())
                .isEqualTo("100")
            assertWithMessage("Intent contains incorrect label")
                .that(node["label"].asText())
                .isEqualTo("Computer Science Orientation")
            assertWithMessage("Intent should not contain description")
                .that(node["description"])
                .isNull()
            assertWithMessage("Intent contains extra fields")
                .that(node.size())
                .isEqualTo(3)
        }
    }

    /** Test CourseActivity UI launched via intent. */
    @Test(timeout = 10000L)
    @Graded(points = 20, friendlyName = "Course View")
    @LazyApplication(LazyApplication.LazyLoad.ON)
    fun test4_CourseView() {
        summariesForUITest.forEach { summaryIndex ->
            // Create the Intent
            val intent = Intent(ApplicationProvider.getApplicationContext(), CourseActivity::class.java).apply {
                putExtra("summary", objectMapper.writeValueAsString(SUMMARIES[summaryIndex]))
            }

            // Start the CourseActivity and ensure that the title and description are shown
            startActivity<CourseActivity>(intent) {
                val course = objectMapper.readTree(COURSES[summaryIndex])
                pause()
                val title =
                    "${course["subject"].asText()} ${course["number"].asText()}: ${course["label"].asText()}"
                onView(withSubstring(title)).check(matches(isDisplayed()))
                onView(withSubstring(course["description"].asText())).check(matches(isDisplayed()))
            }
        }
    }

    companion object {
        // Security manager used to count access to courses.json
        private val coursesReadCountSecurityManager = CoursesReadCountSecurityManager()

        // Run once before any test in this suite is started
        @JvmStatic
        @BeforeClass
        fun beforeClass() {
            // Check Server.kt for publicly visible members
            doCheckServerDesign()
            // Set up logging so that you can see log output during testing
            configureLogging()
            // Force loads to perform initialization before we start counting
            COURSES.size
            SUMMARIES.size
            // Install our security manager that allows counting access to courses.json
            System.setSecurityManager(coursesReadCountSecurityManager)
            // Start the API server
            startServer()
        }

        // Run once after all tests in this suite are completed
        @JvmStatic
        @AfterClass
        fun removeSecurityManager() {
            System.setSecurityManager(null)
        }
    }

    // Run before each test
    @Before
    fun beforeEachTest() {
        // Check for extra reads from courses.json
        // 3 is what's expected based on Robolectric initialization
        assertWithMessage("courses.json should only be accessed during server start")
            .that(coursesReadCountSecurityManager.coursesReadCount)
            .isAtMost(3)
    }

    // Run after each test
    @After
    fun afterEachTest() {
        // Check for extra reads from courses.json
        // 3 is what's expected based on Robolectric initialization
        assertWithMessage("courses.json should only be accessed once during server start")
            .that(coursesReadCountSecurityManager.coursesReadCount)
            .isAtMost(3)
    }
}

private fun doCheckServerDesign() {
    // Check for extra public methods, fields, or inner classes
    val nonPrivateMethodCount = Server::class.java.declaredMethods.count {
        !Modifier.isPrivate(it.modifiers) && !Modifier.isStatic(it.modifiers) && it.name != "dispatch"
    }
    val nonPrivateFieldCount = Server::class.java.declaredFields.count {
        !Modifier.isPrivate(it.modifiers) && it.name != "INSTANCE"
    }
    val nonPrivateClassCount = Server::class.java.declaredClasses.count { !Modifier.isPrivate(it.modifiers) }
    assertWithMessage("Server has visible methods, fields, or classes")
        .that(nonPrivateMethodCount + nonPrivateFieldCount + nonPrivateClassCount)
        .isEqualTo(0)
}

// md5: 1de0d5045f0c12d5857d1e89469a9cc3 // DO NOT REMOVE THIS LINE
